<?php
/**
 * Website: http://sourceforge.net/projects/simplehtmldom/
 * Additional projects: http://sourceforge.net/projects/debugobject/
 * Acknowledge: Jose Solorzano (https://sourceforge.net/projects/php-html/)
 *
 * Licensed under The MIT License
 * See the LICENSE file in the project root for more information.
 *
 * Authors:
 *   S.C. Chen
 *   John Schlick
 *   Rus Carroll
 *   logmanoriginal
 *
 * Contributors:
 *   Yousuke Kumakura
 *   Vadim Voituk
 *   Antcs
 *
 * Version Rev. 1.9.1 (291)
 * 
 * PHP version 5+
 * 
 * @category Helper_Functions
 * @package  PlaceLocalInclude
 * @author   S.C. Chen <me578022@gmail.com>
 * @license  https://mit-license.org/ MIT license
 * @link     https://simplehtmldom.sourceforge.io/
 */

define('HDOM_TYPE_ELEMENT', 1);
define('HDOM_TYPE_COMMENT', 2);
define('HDOM_TYPE_TEXT', 3);
define('HDOM_TYPE_ENDTAG', 4);
define('HDOM_TYPE_ROOT', 5);
define('HDOM_TYPE_UNKNOWN', 6);
define('HDOM_QUOTE_DOUBLE', 0);
define('HDOM_QUOTE_SINGLE', 1);
define('HDOM_QUOTE_NO', 3);
define('HDOM_INFO_BEGIN', 0);
define('HDOM_INFO_END', 1);
define('HDOM_INFO_QUOTE', 2);
define('HDOM_INFO_SPACE', 3);
define('HDOM_INFO_TEXT', 4);
define('HDOM_INFO_INNER', 5);
define('HDOM_INFO_OUTER', 6);
define('HDOM_INFO_ENDSPACE', 7);

defined('DEFAULT_TARGET_CHARSET') || define('DEFAULT_TARGET_CHARSET', 'UTF-8');
defined('DEFAULT_BR_TEXT') || define('DEFAULT_BR_TEXT', "\r\n");
defined('DEFAULT_SPAN_TEXT') || define('DEFAULT_SPAN_TEXT', ' ');
defined('MAX_FILE_SIZE') || define('MAX_FILE_SIZE', 600000);
define('HDOM_SMARTY_AS_TEXT', 1);

/**
 * Get html dom from file
 * $maxlen is defined in the code as PHP_STREAM_COPY_ALL which is defined as -1.
 * 
 * @param string   $url              Name or url of the file
 * @param int      $use_include_path Trigger include path search
 * @param resource $context          Context resource
 * @param int      $offset           Offset of where reading starts
 * @param int      $maxLen           Max length of data to read
 * @param int      $lowercase        Forces lowercase tag matching
 * @param int      $forceTagsClosed  Obsolete
 * @param string   $target_charset   Defines target charset
 * @param int      $stripRN          Removes newlines before parsing
 * @param string   $defaultBRText    Default text for <br> elements
 * @param string   $defaultSpanText  Default text for <span> elements
 * 
 * @return dom
 */
function fileGetHtml(
    $url,
    $use_include_path = false,
    $context = null,
    $offset = 0,
    $maxLen = -1,
    $lowercase = true,
    $forceTagsClosed = true,
    $target_charset = DEFAULT_TARGET_CHARSET,
    $stripRN = true,
    $defaultBRText = DEFAULT_BR_TEXT,
    $defaultSpanText = DEFAULT_SPAN_TEXT
) {
    if ($maxLen <= 0) {
        $maxLen = MAX_FILE_SIZE; 
    }

    $dom = new simpleHtmlDom(
        null,
        $lowercase,
        $forceTagsClosed,
        $target_charset,
        $stripRN,
        $defaultBRText,
        $defaultSpanText
    );

    /**
     * For sourceforge users: uncomment the next line and comment the
     * retrieve_url_contents line 2 lines down if it is not already done.
     */
    $contents = file_get_contents(
        $url,
        $use_include_path,
        $context,
        $offset,
        $maxLen
    );
    // $contents = retrieve_url_contents($url);

    if (empty($contents) || strlen($contents) > $maxLen) {
        $dom->clear();
        return false;
    }

    return $dom->load($contents, $lowercase, $stripRN);
}

/**
 * Get html dom from string
 * 
 * @param string $str             HTML document string
 * @param int    $lowercase       Forces lowercase match of tags
 * @param int    $forceTagsClosed Obsolete
 * @param string $target_charset  Defines target charset
 * @param int    $stripRN         Removes newlines before parsing
 * @param string $defaultBRText   Defines default text for <br> elements
 * @param string $defaultSpanText Defines default text for <span> elements
 * 
 * @return $dom
 */
function strGetHtml(
    $str,
    $lowercase = true,
    $forceTagsClosed = true,
    $target_charset = DEFAULT_TARGET_CHARSET,
    $stripRN = true,
    $defaultBRText = DEFAULT_BR_TEXT,
    $defaultSpanText = DEFAULT_SPAN_TEXT
) {
    $dom = new simpleHtmlDom(
        null,
        $lowercase,
        $forceTagsClosed,
        $target_charset,
        $stripRN,
        $defaultBRText,
        $defaultSpanText
    );

    if (empty($str) || strlen($str) > MAX_FILE_SIZE) {
        $dom->clear();
        return false;
    }

    return $dom->load($str, $lowercase, $stripRN);
}

/**
 * Dump html dom tree
 * 
 * @param object $node      Node to dump
 * @param bool   $show_attr Node attributes
 * @param int    $deep      Depth
 * 
 * @return void
 */
function dumpHtmlTree($node, $show_attr = true, $deep = 0)
{
    $node->dump($node);
}

/**
 * SimpleHtmlDomNode
 *
 * @category Helper_Functions
 * @package  PlaceLocalInclude
 * @author   S.C. Chen <me578022@gmail.com>
 * @license  https://mit-license.org/ MIT license
 * @link     https://simplehtmldom.sourceforge.io/
 */
class SimpleHtmlDomNode
{
    public $nodetype = HDOM_TYPE_TEXT;
    public $tag = 'text';
    public $attr = array();
    public $children = array();
    public $nodes = array();
    public $parent = null;
    public $p_ = array();
    public $tag_start = 0;
    private $_dom = null;

    /**
     * Class constructor
     * 
     * @param dom $_dom New object to create
     * 
     * @return void
     */
    function __construct($_dom)
    {
        $this->_dom = $_dom;
        $_dom->nodes[] = $this;
    }

    /**
     * Class destructor
     * 
     * @return void
     */
    function __destruct()
    {
        $this->clear();
    }

    /**
     *  Root element text
     * 
     * @return string
     */
    function __toString()
    {
        return $this->outertext();
    }

    /**
     * Clean up memory due to php5 circular references memory leak...
     * 
     * @return void
     */
    function clear()
    {
        $this->_dom = null;
        $this->nodes = null;
        $this->parent = null;
        $this->children = null;
    }

    /**
     * Dump node's tree
     * 
     * @param int $show_attr Include attributes in dump
     * @param int $depth     Depth of current element
     * 
     * @return void
     */
    function dump($show_attr = true, $depth = 0)
    {
        echo str_repeat("\t", $depth) . $this->tag;

        if ($show_attr && count($this->attr) > 0) {
            echo '(';
            foreach ($this->attr as $k => $v) {
                echo "[$k]=>\"$v\", ";
            }
            echo ')';
        }

        echo "\n";

        if ($this->nodes) {
            foreach ($this->nodes as $node) {
                $node->dump($show_attr, $depth + 1);
            }
        }
    }

    /**
     * Debugging function to dump a single dom node with a bunch of
     * information about it.
     * 
     * @param int $echo Dump details
     * 
     * @return string $string
     */
    function dumpNode($echo = true)
    {
        $string = $this->tag;

        if (count($this->attr) > 0) {
            $string .= '(';
            foreach ($this->attr as $k => $v) {
                $string .= "[$k]=>\"$v\", ";
            }
            $string .= ')';
        }

        if (count($this->p_) > 0) {
            $string .= ' $p_ (';
            foreach ($this->p_ as $k => $v) {
                if (is_array($v)) {
                    $string .= "[$k]=>(";
                    foreach ($v as $k2 => $v2) {
                        $string .= "[$k2]=>\"$v2\", ";
                    }
                    $string .= ')';
                } else {
                    $string .= "[$k]=>\"$v\", ";
                }
            }
            $string .= ')';
        }

        if (isset($this->text)) {
            $string .= " text: ({$this->text})";
        }

        $string .= ' HDOM_INNER_INFO: ';

        if (isset($node->p_[HDOM_INFO_INNER])) {
            $string .= "'" . $node->p_[HDOM_INFO_INNER] . "'";
        } else {
            $string .= ' NULL ';
        }

        $string .= ' children: ' . count($this->children);
        $string .= ' nodes: ' . count($this->nodes);
        $string .= ' tag_start: ' . $this->tag_start;
        $string .= "\n";

        if ($echo) {
            echo $string;
            return;
        } else {
            return $string;
        }
    }

    /**
     * Returns the parent of node
     * If a node is passed in, it will reset the parent of the current node
     * to that one.
     * 
     * @param object $parent The parent node
     * 
     * @return object
     */
    function parent($parent = null)
    {
        // I am SURE that this doesn't work properly.
        // It fails to unset the current node from it's current parents nodes or
        // children list first.
        if ($parent !== null) {
            $this->parent = $parent;
            $this->parent->nodes[] = $this;
            $this->parent->children[] = $this;
        }

        return $this->parent;
    }

    /**
     * Verify that node has children
     * 
     * @return bool
     */
    function hasChild()
    {
        return !empty($this->children);
    }

    /**
     * Returns children of node
     * 
     * @param int $idx Index of child
     * 
     * @return object
     */
    function children($idx = -1)
    {
        if ($idx === -1) {
            return $this->children;
        }

        if (isset($this->children[$idx])) {
            return $this->children[$idx];
        }

        return null;
    }

    /**
     * Returns the first child of node
     * 
     * @return object
     */
    function firstChild()
    {
        if (count($this->children) > 0) {
            return $this->children[0];
        }
        return null;
    }

    /**
     * Returns the last child of node
     * 
     * @return object
     */
    function lastChild()
    {
        if (count($this->children) > 0) {
            return end($this->children);
        }
        return null;
    }

    /**
     * Returns the next sibling of node
     * 
     * @return object
     */
    function nextSibling()
    {
        if ($this->parent === null) {
            return null;
        }

        $idx = array_search($this, $this->parent->children, true);

        if ($idx !== false && isset($this->parent->children[$idx + 1])) {
            return $this->parent->children[$idx + 1];
        }

        return null;
    }

    /**
     * Returns the previous sibling of node
     * 
     * @return object
     */
    function prevSibling()
    {
        if ($this->parent === null) {
            return null;
        }

        $idx = array_search($this, $this->parent->children, true);

        if ($idx !== false && $idx > 0) {
            return $this->parent->children[$idx - 1];
        }

        return null;
    }

    /**
     * Function to locate a specific ancestor tag in the path to the root.
     * 
     * @param string $tag Object
     * 
     * @return object
     */
    function findAncestorTag($tag)
    {
        global $debug_object;
        if (is_object($debug_object)) {
            $debug_object->debug_log_entry(1); 
        }

        if ($this->parent === null) {
            return null;
        }

        $ancestor = $this->parent;

        while (!is_null($ancestor)) {
            if (is_object($debug_object)) {
                $debug_object->debug_log(2, 'Current tag is: ' . $ancestor->tag);
            }

            if ($ancestor->tag === $tag) {
                break;
            }

            $ancestor = $ancestor->parent;
        }

        return $ancestor;
    }

    /**
     * Get dom node's inner html
     * 
     * @return string
     */
    function innertext()
    {
        if (isset($this->p_[HDOM_INFO_INNER])) {
            return $this->p_[HDOM_INFO_INNER];
        }

        if (isset($this->p_[HDOM_INFO_TEXT])) {
            return $this->_dom->restoreNoise($this->p_[HDOM_INFO_TEXT]);
        }

        $ret = '';

        foreach ($this->nodes as $n) {
            $ret .= $n->outertext();
        }

        return $ret;
    }

    /**
     * Get dom node's outer text (with tag)
     * 
     * @return string
     */
    function outertext()
    {
        global $debug_object;

        if (is_object($debug_object)) {
            $text = '';

            if ($this->tag === 'text') {
                if (!empty($this->text)) {
                    $text = ' with text: ' . $this->text;
                }
            }

            $debug_object->debug_log(1, 'Innertext of tag: ' . $this->tag . $text);
        }

        if ($this->tag === 'root') {
            return $this->innertext();
        }

        // todo: What is the use of this callback? Remove?
        if ($this->_dom && $this->_dom->callback !== null) {
            call_user_func_array($this->_dom->callback, array($this));
        }

        if (isset($this->p_[HDOM_INFO_OUTER])) {
            return $this->p_[HDOM_INFO_OUTER];
        }

        if (isset($this->p_[HDOM_INFO_TEXT])) {
            return $this->_dom->restoreNoise($this->p_[HDOM_INFO_TEXT]);
        }

        $ret = '';

        if ($this->_dom && $this->_dom->nodes[$this->p_[HDOM_INFO_BEGIN]]) {
            $ret = $this->_dom->nodes[$this->p_[HDOM_INFO_BEGIN]]->makeup();
        }

        if (isset($this->p_[HDOM_INFO_INNER])) {
            // todo: <br> should either never have HDOM_INFO_INNER or always
            if ($this->tag !== 'br') {
                $ret .= $this->p_[HDOM_INFO_INNER];
            }
        } elseif ($this->nodes) {
            foreach ($this->nodes as $n) {
                $ret .= $this->convertText($n->outertext());
            }
        }

        if (isset($this->p_[HDOM_INFO_END]) && $this->p_[HDOM_INFO_END] != 0) {
            $ret .= '</' . $this->tag . '>';
        }

        return $ret;
    }

    /**
     * Get dom node's plain text
     * 
     * @return string
     */
    function text()
    {
        if (isset($this->p_[HDOM_INFO_INNER])) {
            return $this->p_[HDOM_INFO_INNER];
        }

        switch ($this->nodetype) {
        case HDOM_TYPE_TEXT: 
            return $this->_dom->restoreNoise($this->p_[HDOM_INFO_TEXT]);
        case HDOM_TYPE_COMMENT: 
            return '';
        case HDOM_TYPE_UNKNOWN: 
            return '';
        }

        if (strcasecmp($this->tag, 'script') === 0) {
            return ''; 
        }
        if (strcasecmp($this->tag, 'style') === 0) {
            return ''; 
        }

        $ret = '';

        // In rare cases, (always node type 1 or HDOM_TYPE_ELEMENT - observed
        // for some span tags, and some p tags) $this->nodes is set to NULL.
        // NOTE: This indicates that there is a problem where it's set to NULL
        // without a clear happening.
        // WHY is this happening?
        if (!is_null($this->nodes)) {
            foreach ($this->nodes as $n) {
                // Start paragraph after a blank line
                if ($n->tag === 'p') {
                    $ret = trim($ret) . "\n\n";
                }

                $ret .= $this->convertText($n->text());

                // If this node is a span... add a space at the end of it so
                // multiple spans don't run into each other.  This is plaintext
                // after all.
                if ($n->tag === 'span') {
                    $ret .= $this->_dom->default_span_text;
                }
            }
        }
        return $ret;
    }

    /**
     * Xml text
     * 
     * @return string
     */
    function xmltext()
    {
        $ret = $this->innertext();
        $ret = str_ireplace('<![CDATA[', '', $ret);
        $ret = str_replace(']]>', '', $ret);
        return $ret;
    }

    /**
     * Build node's text with tag
     * 
     * @return string
     */
    function makeup()
    {
        // text, comment, unknown
        if (isset($this->p_[HDOM_INFO_TEXT])) {
            return $this->_dom->restoreNoise($this->p_[HDOM_INFO_TEXT]);
        }

        $ret = '<' . $this->tag;
        $i = -1;

        foreach ($this->attr as $key => $val) {
            ++$i;

            // skip removed attribute
            if ($val === null || $val === false) {
                continue; 
            }

            $ret .= $this->p_[HDOM_INFO_SPACE][$i][0];

            //no value attr: nowrap, checked selected...
            if ($val === true) {
                $ret .= $key;
            } else {
                switch ($this->p_[HDOM_INFO_QUOTE][$i])
                {
                case HDOM_QUOTE_DOUBLE: $quote = '"'; 
                    break;
                case HDOM_QUOTE_SINGLE: $quote = '\''; 
                    break;
                default: $quote = '';
                }

                $ret .= $key
                . $this->p_[HDOM_INFO_SPACE][$i][1]
                . '='
                . $this->p_[HDOM_INFO_SPACE][$i][2]
                . $quote
                . $val
                . $quote;
            }
        }

        $ret = $this->_dom->restoreNoise($ret);
        return $ret . $this->p_[HDOM_INFO_ENDSPACE] . '>';
    }

    /**
     * Find elements by css selector
     * 
     * @param string $selector  CSS selector
     * @param int    $idx       Index of element to return
     * @param bool   $lowercase Makes tags case insensitive
     * 
     * @return object
     */
    function find($selector, $idx = null, $lowercase = false)
    {
        $selectors = $this->parseSelector($selector);
        if (($count = count($selectors)) === 0) {
            return array(); 
        }
        $found_keys = array();

        // find each selector
        for ($c = 0; $c < $count; ++$c) {
            // The change on the below line was documented on the sourceforge
            // code tracker id 2788009
            // used to be: if (($levle=count($selectors[0]))===0) return array();
            if (($levle = count($selectors[$c])) === 0) {
                return array(); 
            }
            if (!isset($this->p_[HDOM_INFO_BEGIN])) {
                return array(); 
            }

            $head = array($this->p_[HDOM_INFO_BEGIN] => 1);
            $cmd = ' '; // Combinator

            // handle descendant selectors, no recursive!
            for ($l = 0; $l < $levle; ++$l) {
                $ret = array();

                foreach ($head as $k => $v) {
                    $n = ($k === -1) ? $this->_dom->root : $this->_dom->nodes[$k];
                    //PaperG - Pass this optional parameter on to the seek function.
                    $n->seek($selectors[$c][$l], $ret, $cmd, $lowercase);
                }

                $head = $ret;
                $cmd = $selectors[$c][$l][4]; // Next Combinator
            }

            foreach ($head as $k => $v) {
                if (!isset($found_keys[$k])) {
                    $found_keys[$k] = 1;
                }
            }
        }

        // sort keys
        ksort($found_keys);

        $found = array();
        foreach ($found_keys as $k => $v) {
            $found[] = $this->_dom->nodes[$k];
        }

        // return nth-element or array
        if (is_null($idx)) {
            return $found; 
        } elseif ($idx < 0) {
            $idx = count($found) + $idx; 
        }
        return (isset($found[$idx])) ? $found[$idx] : null;
    }

    /**
     * Seek fore given conditions
     * 
     * @param string $selector   Current selector
     * @param array  $ret        Starting point (previous return value)
     * @param string $parent_cmd Combinator used before current selector
     * @param bool   $lowercase  Matches tag names case insensitive
     * 
     * @return void
     */
    protected function seek($selector, &$ret, $parent_cmd, $lowercase = false)
    {
        global $debug_object;
        if (is_object($debug_object)) {
            $debug_object->debug_log_entry(1); 
        }

        list($tag, $id, $class, $attributes, $cmb) = $selector;
        $nodes = array();

        if ($parent_cmd === ' ') { // Descendant Combinator
            // Find parent closing tag if the current element doesn't have a closing
            // tag (i.e. void element)
            $end = (!empty($this->p_[HDOM_INFO_END])) ? $this->p_[HDOM_INFO_END] : 0;
            if ($end == 0) {
                $parent = $this->parent;
                while (!isset($parent->p_[HDOM_INFO_END]) && $parent !== null) {
                    $end -= 1;
                    $parent = $parent->parent;
                }
                $end += $parent->p_[HDOM_INFO_END];
            }

            // Get list of target nodes
            $nodes_start = $this->p_[HDOM_INFO_BEGIN] + 1;
            $nodes_count = $end - $nodes_start;
            $nodes = array_slice(
                $this->_dom->nodes, $nodes_start, $nodes_count, true
            );
        } elseif ($parent_cmd === '>') { // Child Combinator
            $nodes = $this->children;
        } elseif ($parent_cmd === '+'
            && $this->parent
            && in_array($this, $this->parent->children)
        ) { // Next-Sibling Combinator
            $index = array_search($this, $this->parent->children, true) + 1;
            if ($index < count($this->parent->children)) {
                $nodes[] = $this->parent->children[$index];
            }
        } elseif ($parent_cmd === '~'
            && $this->parent
            && in_array($this, $this->parent->children)
        ) { // Subsequent Sibling Combinator
            $index = array_search($this, $this->parent->children, true);
            $nodes = array_slice($this->parent->children, $index);
        }

        // Go throgh each element starting at this element until the end tag
        // Note: If this element is a void tag, any previous void element is
        // skipped.
        foreach ($nodes as $node) {
            $pass = true;

            // Skip root nodes
            if (!$node->parent) {
                $pass = false;
            }

            // Handle 'text' selector
            if ($pass && $tag === 'text' && $node->tag === 'text') {
                $ret[array_search($node, $this->_dom->nodes, true)] = 1;
                unset($node);
                continue;
            }

            // Skip if node isn't a child node (i.e. text nodes)
            if ($pass && !in_array($node, $node->parent->children, true)) {
                $pass = false;
            }

            // Skip if tag doesn't match
            if ($pass && $tag !== '' && $tag !== $node->tag && $tag !== '*') {
                $pass = false;
            }

            // Skip if ID doesn't exist
            if ($pass && $id !== '' && !isset($node->attr['id'])) {
                $pass = false;
            }

            // Check if ID matches
            if ($pass && $id !== '' && isset($node->attr['id'])) {
                // Note: Only consider the first ID (as browsers do)
                $node_id = explode(' ', trim($node->attr['id']))[0];

                if ($id !== $node_id) {
                    $pass = false; 
                }
            }

            // Check if all class(es) exist
            if ($pass && $class !== '' && is_array($class) && !empty($class)) {
                if (isset($node->attr['class'])) {
                    $node_classes = explode(' ', $node->attr['class']);

                    if ($lowercase) {
                        $node_classes = array_map('strtolower', $node_classes);
                    }

                    foreach ($class as $c) {
                        if (!in_array($c, $node_classes)) {
                            $pass = false;
                            break;
                        }
                    }
                } else {
                    $pass = false;
                }
            }

            // Check attributes
            if ($pass
                && $attributes !== ''
                && is_array($attributes)
                && !empty($attributes)
            ) {
                foreach ($attributes as $a) {
                    list (
                    $att_name,
                    $att_expr,
                    $att_val,
                    $att_inv,
                    $att_case_sensitivity
                    ) = $a;

                    // Handle indexing attributes (i.e. "[2]")
                    /**
                     * Note: This is not supported by the CSS Standard but adds
                     * the ability to select items compatible to XPath (i.e.
                     * the 3rd element within it's parent).
                     *
                     * Note: This doesn't conflict with the CSS Standard which
                     * doesn't work on numeric attributes anyway.
                     */
                    if (is_numeric($att_name)
                        && $att_expr === ''
                        && $att_val === ''
                    ) {
                             $count = 0;

                             // Find index of current element in parent
                        foreach ($node->parent->children as $c) {
                            if ($c->tag === $node->tag) {
                                ++$count;
                            }
                            if ($c === $node) {
                                break;
                            }
                        }

                        // If this is the correct node, continue with next
                        // attribute
                        if ($count === (int)$att_name) {
                            continue;
                        }
                    }

                    // Check attribute availability
                    if ($att_inv) { // Attribute should NOT be set
                        if (isset($node->attr[$att_name])) {
                              $pass = false;
                              break;
                        }
                    } else { // Attribute should be set
                        // todo: "plaintext" is not a valid CSS selector!
                        if ($att_name !== 'plaintext'
                            && !isset($node->attr[$att_name])
                        ) {
                            $pass = false;
                            break;
                        }
                    }

                    // Continue with next attribute if expression isn't defined
                    if ($att_expr === '') {
                        continue;
                    }

                    // If they have told us that this is a "plaintext"
                    // search then we want the plaintext of the node - right?
                    // todo "plaintext" is not a valid CSS selector!
                    if ($att_name === 'plaintext') {
                          $nodeKeyValue = $node->text();
                    } else {
                        $nodeKeyValue = $node->attr[$att_name];
                    }

                    if (is_object($debug_object)) {
                        $debug_object->debug_log(
                            2,
                            'testing node: '
                            . $node->tag
                            . ' for attribute: '
                            . $att_name
                            . $att_expr
                            . $att_val
                            . ' where nodes value is: '
                            . $nodeKeyValue
                        );
                    }

                    // If lowercase is set, do a case insensitive test of
                    // the value of the selector.
                    if ($lowercase) {
                        $check = $this->match(
                            $att_expr,
                            strtolower($att_val),
                            strtolower($nodeKeyValue),
                            $att_case_sensitivity
                        );
                    } else {
                        $check = $this->match(
                            $att_expr,
                            $att_val,
                            $nodeKeyValue,
                            $att_case_sensitivity
                        );
                    }

                    if (is_object($debug_object)) {
                        $debug_object->debug_log(
                            2,
                            'after match: '
                            . ($check ? 'true' : 'false')
                        );
                    }

                    if (!$check) {
                        $pass = false;
                        break;
                    }
                }
            }

            // Found a match. Add to list and clear node
            if ($pass) {
                $ret[$node->p_[HDOM_INFO_BEGIN]] = 1;
            }
            unset($node);
        }
        // It's passed by reference so this is actually what this function returns.
        if (is_object($debug_object)) {
            $debug_object->debug_log(1, 'EXIT - ret: ', $ret);
        }
    }

    /**
     * Match specified attribute selector
     * 
     * @param string $exp              Expression
     * @param string $pattern          Pattern
     * @param string $value            Value
     * @param string $case_sensitivity Case sensitivity
     * 
     * @return bool
     */
    protected function match($exp, $pattern, $value, $case_sensitivity)
    {
        global $debug_object;
        if (is_object($debug_object)) {
            $debug_object->debug_log_entry(1);
        }

        if ($case_sensitivity === 'i') {
            $pattern = strtolower($pattern);
            $value = strtolower($value);
        }

        switch ($exp) {
        case '=':
            return ($value === $pattern);
        case '!=':
            return ($value !== $pattern);
        case '^=':
            return preg_match('/^' . preg_quote($pattern, '/') . '/', $value);
        case '$=':
            return preg_match('/' . preg_quote($pattern, '/') . '$/', $value);
        case '*=':
            return preg_match('/' . preg_quote($pattern, '/') . '/', $value);
        case '|=':
            /**
             * [att|=val]
             *
             * Represents an element with the att attribute, its value
             * either being exactly "val" or beginning with "val"
             * immediately followed by "-" (U+002D).
             */
            return strpos($value, $pattern) === 0;
        case '~=':
            /**
             * [att~=val]
             *
             * Represents an element with the att attribute whose value is a
             * whitespace-separated list of words, one of which is exactly
             * "val". If "val" contains whitespace, it will never represent
             * anything (since the words are separated by spaces). Also if
             * "val" is the empty string, it will never represent anything.
             */
            return in_array($pattern, explode(' ', trim($value)), true);
        }
        return false;
    }

    /**
     * Parse CSS selector
     * 
     * @param string $selector_string The selector string
     * 
     * @return array
     */
    protected function parseSelector($selector_string)
    {
        global $debug_object;
        if (is_object($debug_object)) {
            $debug_object->debug_log_entry(1); 
        }

        /**
         * Pattern of CSS selectors, modified from mootools (https://mootools.net/)
         *
         * Paperg: Add the colon to the attribute, so that it properly finds
         * <tag attr:ibute="something" > like google does.
         *
         * Note: if you try to look at this attribute, you MUST use getAttribute
         * since $dom->x:y will fail the php syntax check.
         *
         * Notice the \[ starting the attribute? and the @? following? This
         * implies that an attribute can begin with an @ sign that is not
         * captured. This implies that an html attribute specifier may start
         * with an @ sign that is NOT captured by the expression. Farther study
         * is required to determine of this should be documented or removed.
         *
         * Matches selectors in this order:
         *
         * [0] - full match
         *
         * [1] - tag name
         *     ([\w:\*-]*)
         *     Matches the tag name consisting of zero or more words, colons,
         *     asterisks and hyphens.
         *
         * [2] - id name
         *     (?:\#([\w-]+))
         *     Optionally matches a id name, consisting of an "#" followed by
         *     the id name (one or more words and hyphens).
         *
         * [3] - class names (including dots)
         *     (?:\.([\w\.-]+))?
         *     Optionally matches a list of classs, consisting of an "."
         *     followed by the class name (one or more words and hyphens)
         *     where multiple classes can be chained (i.e. ".foo.bar.baz")
         *
         * [4] - attributes
         *     ((?:\[@?(?:!?[\w:-]+)(?:(?:[!*^$|~]?=)[\"']?(?:.*?)[\"']?)?(?:\s*?(?:[iIsS])?)?\])+)?
         *     Optionally matches the attributes list
         *
         * [5] - separator
         *     ([\/, >+~]+)
         *     Matches the selector list separator
         */
     // phpcs:ignore Generic.Files.LineLength
        $pattern = "/([\w:\*-]*)(?:\#([\w-]+))?(?:|\.([\w\.-]+))?((?:\[@?(?:!?[\w:-]+)(?:(?:[!*^$|~]?=)[\"']?(?:.*?)[\"']?)?(?:\s*?(?:[iIsS])?)?\])+)?([\/, >+~]+)/is";

        preg_match_all(
            $pattern,
            trim($selector_string) . ' ', // Add final ' ' as pseudo separator
            $matches,
            PREG_SET_ORDER
        );

        if (is_object($debug_object)) {
            $debug_object->debug_log(2, 'Matches Array: ', $matches);
        }

        $selectors = array();
        $result = array();

        foreach ($matches as $m) {
            $m[0] = trim($m[0]);

            // Skip NoOps
            if ($m[0] === '' || $m[0] === '/' || $m[0] === '//') {
                continue; 
            }

            // Convert to lowercase
            if ($this->_dom->lowercase) {
                $m[1] = strtolower($m[1]);
            }

            // Extract classes
            if ($m[3] !== '') {
                $m[3] = explode('.', $m[3]); 
            }

            /* Extract attributes (pattern based on the pattern above!)

            * [0] - full match
            * [1] - attribute name
            * [2] - attribute expression
            * [3] - attribute value
            * [4] - case sensitivity
            *
            * Note: Attributes can be negated with a "!" prefix to their name
            */
            if ($m[4] !== '') {
                preg_match_all(
                    "/\[@?(!?[\w:-]+)(?:([!*^$|~]?=)[\"']?(.*?)[\"']?)?(?:\s+?"
                        ."([iIsS])?)?\]/is",
                    trim($m[4]),
                    $attributes,
                    PREG_SET_ORDER
                );

                // Replace element by array
                $m[4] = array();

                foreach ($attributes as $att) {
                       // Skip empty matches
                    if (trim($att[0]) === '') {
                        continue; 
                    }

                       $inverted = (isset($att[1][0]) && $att[1][0] === '!');
                       $m[4][] = array(
                        $inverted ? substr($att[1], 1) : $att[1], // Name
                        (isset($att[2])) ? $att[2] : '', // Expression
                        (isset($att[3])) ? $att[3] : '', // Value
                        $inverted, // Inverted Flag
                        (isset($att[4])) ?
                            strtolower($att[4]) : '', // Case-Sensitivity
                       );
                }
            }

            // Sanitize Separator
            if ($m[5] !== '' && trim($m[5]) === '') { // Descendant Separator
                $m[5] = ' ';
            } else { // Other Separator
                $m[5] = trim($m[5]);
            }

            // Clear Separator if it's a Selector List
            if ($is_list = ($m[5] === ',')) {
                $m[5] = ''; 
            }

            // Remove full match before adding to results
            array_shift($m);
            $result[] = $m;

            if ($is_list) { // Selector List
                $selectors[] = $result;
                $result = array();
            }
        }

        if (count($result) > 0) {
            $selectors[] = $result; 
        }
        return $selectors;
    }

    /**
     * Return attribut value
     * 
     * @param string $name Attribut name
     * 
     * @return mixed
     */
    function __get($name)
    {
        if (isset($this->attr[$name])) {
            return $this->convertText($this->attr[$name]);
        }
        switch ($name) {
        case 'outertext': 
            return $this->outertext();
        case 'innertext': 
            return $this->innertext();
        case 'plaintext': 
            return $this->text();
        case 'xmltext': 
            return $this->xmltext();
        default: 
            return array_key_exists($name, $this->attr);
        }
    }

    /**
     * Set attribute value
     * 
     * @param string $name  Attribute name
     * @param mixed  $value Value to set
     * 
     * @return mixed
     */
    function __set($name, $value)
    {
        global $debug_object;
        if (is_object($debug_object)) {
            $debug_object->debug_log_entry(1); 
        }

        switch ($name) {
        case 'outertext': 
            return $this->p_[HDOM_INFO_OUTER] = $value;
        case 'innertext':
            if (isset($this->p_[HDOM_INFO_TEXT])) {
                 return $this->p_[HDOM_INFO_TEXT] = $value;
            }
            return $this->p_[HDOM_INFO_INNER] = $value;
        }

        if (!isset($this->attr[$name])) {
            $this->p_[HDOM_INFO_SPACE][] = array(' ', '', '');
            $this->p_[HDOM_INFO_QUOTE][] = HDOM_QUOTE_DOUBLE;
        }

        $this->attr[$name] = $value;
    }

    /**
     * Is attribute set
     * 
     * @param string $name Attribute name
     * 
     * @return bool
     */
    function __isset($name)
    {
        switch ($name) {
        case 'outertext': 
            return true;
        case 'innertext': 
            return true;
        case 'plaintext': 
            return true;
        }
        //no value attr: nowrap, checked selected...
        return (array_key_exists($name, $this->attr)) ?
            true : isset($this->attr[$name]);
    }

    /**
     * Remove attribute
     * 
     * @param string $name Attribute name
     * 
     * @return void
     */
    function __unset($name)
    {
        if (isset($this->attr[$name])) {
            unset($this->attr[$name]); 
        }
    }

    /**
     * Convert text
     * 
     * @param string $text Text to convert
     * 
     * @return string
     */
    function convertText($text)
    {
        global $debug_object;
        if (is_object($debug_object)) {
            $debug_object->debug_log_entry(1); 
        }

        $converted_text = $text;

        $sourceCharset = '';
        $targetCharset = '';

        if ($this->_dom) {
            $sourceCharset = strtoupper($this->_dom->_charset ?? '');
            $targetCharset = strtoupper($this->_dom->_target_charset ?? '');
        }

        if (is_object($debug_object)) {
            $debug_object->debug_log(
                3,
                'source charset: '
                . $sourceCharset
                . ' target charaset: '
                . $targetCharset
            );
        }

        if (!empty($sourceCharset)
            && !empty($targetCharset)
            && (strcasecmp($sourceCharset, $targetCharset) != 0)
        ) {
            // Check if the reported encoding could have been incorrect
            // and the text is actually already UTF-8
            if ((strcasecmp($targetCharset, 'UTF-8') == 0)
                && ($this->isUtf8($text))
            ) {
                $converted_text = $text;
            } else {
                $converted_text = iconv($sourceCharset, $targetCharset, $text);
            }
        }

        // Lets make sure that we don't have that silly BOM issue with any
        // of the utf-8 text we output.
        if ($targetCharset === 'UTF-8') {
            if (substr($converted_text, 0, 3) === "\xef\xbb\xbf") {
                $converted_text = substr($converted_text, 3);
            }

            if (substr($converted_text, -3) === "\xef\xbb\xbf") {
                $converted_text = substr($converted_text, 0, -3);
            }
        }

        return $converted_text;
    }

    /**
     * Test for utf-8
     * 
     * @param string $str String to test
     * 
     * @return bool
     */
    static function isUtf8($str)
    {
        $c = 0; $b = 0;
        $bits = 0;
        $len = strlen($str);
        for ( $i = 0; $i < $len; $i++) {
            $c = ord($str[$i]);
            if ($c > 128) {
                if (($c >= 254)) {
                    return false; 
                } elseif ($c >= 252) {
                    $bits = 6; 
                } elseif ($c >= 248) {
                    $bits = 5; 
                } elseif ($c >= 240) {
                    $bits = 4; 
                } elseif ($c >= 224) {
                    $bits = 3; 
                } elseif ($c >= 192) {
                    $bits = 2; 
                } else {
                    return false; 
                }
                if (($i + $bits) > $len) {
                    return false; 
                }
                while ($bits > 1) {
                    $i++;
                    $b = ord($str[$i]);
                    if ($b < 128 || $b > 191) {
                        return false; 
                    }
                    $bits--;
                }
            }
        }
        return true;
    }

    /**
     * Get display size
     * 
     * @return mixed
     */
    function getDisplaySize()
    {
        global $debug_object;

        $width = -1;
        $height = -1;

        if ($this->tag !== 'img') {
            return false;
        }

        // See if there is aheight or width attribute in the tag itself.
        if (isset($this->attr['width'])) {
            $width = $this->attr['width'];
        }

        if (isset($this->attr['height'])) {
            $height = $this->attr['height'];
        }

        // Now look for an inline style.
        if (isset($this->attr['style'])) {
            // Thanks to user gnarf from stackoverflow for this regular expression.
            $attributes = array();

            preg_match_all(
                '/([\w-]+)\s*:\s*([^;]+)\s*;?/',
                $this->attr['style'],
                $matches,
                PREG_SET_ORDER
            );

            foreach ($matches as $match) {
                $attributes[$match[1]] = $match[2];
            }

            // If there is a width in the style attributes:
            if (isset($attributes['width']) && $width == -1) {
                // check that the last two characters are px (pixels)
                if (strtolower(substr($attributes['width'], -2)) === 'px') {
                    $proposed_width = substr($attributes['width'], 0, -2);
                    // Now make sure that it's an integer and not something stupid.
                    if (filter_var($proposed_width, FILTER_VALIDATE_INT)) {
                        $width = $proposed_width;
                    }
                }
            }

            // If there is a width in the style attributes:
            if (isset($attributes['height']) && $height == -1) {
                // check that the last two characters are px (pixels)
                if (strtolower(substr($attributes['height'], -2)) == 'px') {
                    $proposed_height = substr($attributes['height'], 0, -2);
                    // Now make sure that it's an integer and not something stupid.
                    if (filter_var($proposed_height, FILTER_VALIDATE_INT)) {
                        $height = $proposed_height;
                    }
                }
            }

        }

        // Future enhancement:
        // Look in the tag to see if there is a class or id specified that has
        // a height or width attribute to it.

        // Far future enhancement
        // Look at all the parent tags of this image to see if they specify a
        // class or id that has an img selector that specifies a height or width
        // Note that in this case, the class or id will have the img subselector
        // for it to apply to the image.

        // ridiculously far future development
        // If the class or id is specified in a SEPARATE css file thats not on
        // the page, go get it and do what we were just doing for the ones on
        // the page.

        $result = array(
        'height' => $height,
        'width' => $width
        );

        return $result;
    }

    /**
     * Save node
     * 
     * @param string $filepath File path
     * 
     * @return string
     */
    function save($filepath = '')
    {
        $ret = $this->outertext();

        if ($filepath !== '') {
            file_put_contents($filepath, $ret, LOCK_EX);
        }

        return $ret;
    }

    /**
     * Add class
     * 
     * @param mixed $class Class to add
     * 
     * @return void
     */
    function addClass($class)
    {
        if (is_string($class)) {
            $class = explode(' ', $class);
        }

        if (is_array($class)) {
            foreach ($class as $c) {
                if (isset($this->class)) {
                    if ($this->hasClass($c)) {
                        continue;
                    } else {
                        $this->class .= ' ' . $c;
                    }
                } else {
                    $this->class = $c;
                }
            }
        } else {
            if (is_object($debug_object)) {
                $debug_object->debug_log(2, 'Invalid type: ', gettype($class));
            }
        }
    }

    /**
     * Has class
     * 
     * @param string $class Class name
     * 
     * @return bool
     */
    function hasClass($class)
    {
        if (is_string($class)) {
            if (isset($this->class)) {
                return in_array($class, explode(' ', $this->class), true);
            }
        } else {
            if (is_object($debug_object)) {
                $debug_object->debug_log(2, 'Invalid type: ', gettype($class));
            }
        }

        return false;
    }

    /**
     * Remove class
     * 
     * @param mixed $class Class name
     * 
     * @return void
     */
    function removeClass($class = null)
    {
        if (!isset($this->class)) {
            return;
        }

        if (is_null($class)) {
            $this->removeAttribute('class');
            return;
        }

        if (is_string($class)) {
            $class = explode(' ', $class);
        }

        if (is_array($class)) {
            $class = array_diff(explode(' ', $this->class), $class);
            if (empty($class)) {
                $this->removeAttribute('class');
            } else {
                $this->class = implode(' ', $class);
            }
        }
    }

    /**
     * Get all attributes
     * 
     * @return array
     */
    function getAllAttributes()
    {
        return $this->attr;
    }

    /**
     * Get attribute
     * 
     * @param string $name Attribute name
     * 
     * @return mixed
     */
    function getAttribute($name)
    {
        return $this->__get($name);
    }

    /**
     * Set attribute
     * 
     * @param string $name  Attribute name
     * @param string $value Attribute value
     * 
     * @return void
     */
    function setAttribute($name, $value)
    {
        $this->__set($name, $value);
    }

    /**
     * Has attribute
     * 
     * @param string $name Attribute name
     * 
     * @return bool
     */
    function hasAttribute($name)
    {
        return $this->__isset($name);
    }

    /**
     * Remove attribute
     * 
     * @param string $name Attribute name
     * 
     * @return void
     */
    function removeAttribute($name)
    {
        $this->__set($name, null);
    }

    /**
     * Remove curreent node
     * 
     * @return void
     */
    function remove()
    {
        if ($this->parent) {
            $this->parent->removeChild($this);
        }
    }

    /**
     * Remove child
     * 
     * @param object $node Node to remove
     * 
     * @return void
     */
    function removeChild($node)
    {
        $nidx = array_search($node, $this->nodes, true);
        $cidx = array_search($node, $this->children, true);
        $didx = array_search($node, $this->_dom->nodes, true);

        if ($nidx !== false && $cidx !== false && $didx !== false) {

            foreach ($node->children as $child) {
                $node->removeChild($child);
            }

            foreach ($node->nodes as $entity) {
                $enidx = array_search($entity, $node->nodes, true);
                $edidx = array_search($entity, $node->_dom->nodes, true);

                if ($enidx !== false && $edidx !== false) {
                    unset($node->nodes[$enidx]);
                    unset($node->_dom->nodes[$edidx]);
                }
            }

            unset($this->nodes[$nidx]);
            unset($this->children[$cidx]);
            unset($this->_dom->nodes[$didx]);

            $node->clear();

        }
    }

    /**
     * Get element by id
     * 
     * @param string $id Element nid
     * 
     * @return object
     */
    function getElementById($id)
    {
        return $this->find("#$id", 0);
    }

    /**
     * Get elements by id
     * 
     * @param string $id  Element id
     * @param int    $idx Element at index
     * 
     * @return object
     */
    function getElementsById($id, $idx = null)
    {
        return $this->find("#$id", $idx);
    }

    /**
     * Get element by tag name
     * 
     * @param string $name Element tag name
     * 
     * @return object
     */
    function getElementByTagName($name)
    {
        return $this->find($name, 0);
    }

    /**
     * Get elements by tag name
     * 
     * @param string $name Element tag name
     * @param int    $idx  Element at index
     * 
     * @return object
     */
    function getElementsByTagName($name, $idx = null)
    {
        return $this->find($name, $idx);
    }

    /**
     * Parent node
     * 
     * @return object
     */
    function parentNode()
    {
        return $this->parent();
    }

    /**
     * Child nodes
     * 
     * @param int $idx Node index
     * 
     * @return mixed
     */
    function childNodes($idx = -1)
    {
        return $this->children($idx);
    }

    //prevSibling likely intended --- would be duplicate
    //function previousSibling()
    //{
    //    return $this->prevSibling();
    //}


    /**
     * Has child wrapper
     * 
     * @return bool
     */
    function hasChildNodes()
    {
        return $this->hasChild();
    }

    /**
     * Node name
     * 
     * @return string
     */
    function nodeName()
    {
        return $this->tag;
    }

    /**
     * Append child
     * 
     * @param object $node Object type
     * 
     * @return object
     */
    function appendChild($node)
    {
        $node->parent($this);
        return $node;
    }

}

/**
 * SimpleHtmlDom
 *
 * @category Helper_Functions
 * @package  PlaceLocalInclude
 * @author   S.C. Chen <me578022@gmail.com>
 * @license  https://mit-license.org/ MIT license
 * @link     https://simplehtmldom.sourceforge.io/
 */
class SimpleHtmlDom
{
    public $root = null;
    public $nodes = array();
    public $callback = null;
    public $lowercase = false;
    public $original_size;
    public $size;

    protected $pos;
    protected $doc;
    protected $char;

    protected $cursor;
    protected $parent;
    protected $noise = array();
    protected $token_blank = " \t\r\n";
    protected $token_equal = ' =/>';
    protected $token_slash = " />\r\n\t";
    protected $token_attr = ' >';

    public $p_charset = '';
    public $p_target_charset = '';

    protected $default_br_text = '';

    public $default_span_text = '';

    protected $self_closing_tags = array(
    'area' => 1,
    'base' => 1,
    'br' => 1,
    'col' => 1,
    'embed' => 1,
    'hr' => 1,
    'img' => 1,
    'input' => 1,
    'link' => 1,
    'meta' => 1,
    'param' => 1,
    'source' => 1,
    'track' => 1,
    'wbr' => 1
    );
    protected $block_tags = array(
    'body' => 1,
    'div' => 1,
    'form' => 1,
    'root' => 1,
    'span' => 1,
    'table' => 1
    );
    protected $optional_closing_tags = array(
    // Not optional, see
    // https://www.w3.org/TR/html/textlevel-semantics.html#the-b-element
    'b' => array('b' => 1),
    'dd' => array('dd' => 1, 'dt' => 1),
    // Not optional, see
    // https://www.w3.org/TR/html/grouping-content.html#the-dl-element
    'dl' => array('dd' => 1, 'dt' => 1),
    'dt' => array('dd' => 1, 'dt' => 1),
    'li' => array('li' => 1),
    'optgroup' => array('optgroup' => 1, 'option' => 1),
    'option' => array('optgroup' => 1, 'option' => 1),
    'p' => array('p' => 1),
    'rp' => array('rp' => 1, 'rt' => 1),
    'rt' => array('rp' => 1, 'rt' => 1),
    'td' => array('td' => 1, 'th' => 1),
    'th' => array('td' => 1, 'th' => 1),
    'tr' => array('td' => 1, 'th' => 1, 'tr' => 1),
    );

    /**
     * Class constructor
     * 
     * @param string $str              HTML document string
     * @param bool   $lowercase        Tag names in lowercase
     * @param bool   $forceTagsClosed  Inside block tags forcefully closed
     * @param string $p_target_charset Target character set
     * @param bool   $stripRN          Newline characters replaced by whitespace
     * @param string $defaultBRText    Default text for <br> elements
     * @param string $defaultSpanText  Default text for <span> elements
     * @param int    $options          Additional options for parser
     * 
     * @return object
     */
    function __construct(
        $str = null,
        $lowercase = true,
        $forceTagsClosed = true,
        $p_target_charset = DEFAULT_TARGET_CHARSET,
        $stripRN = true,
        $defaultBRText = DEFAULT_BR_TEXT,
        $defaultSpanText = DEFAULT_SPAN_TEXT,
        $options = 0
    ) {
        if ($str) {
            if (preg_match('/^http:\/\//i', $str) || is_file($str)) {
                $this->loadFile($str);
            } else {
                $this->load(
                    $str,
                    $lowercase,
                    $stripRN,
                    $defaultBRText,
                    $defaultSpanText,
                    $options
                );
            }
        }
        // Forcing tags to be closed implies that we don't trust the html, but
        // it can lead to parsing errors if we SHOULD trust the html.
        if (!$forceTagsClosed) {
            $this->optional_closing_array = array();
        }

        $this->p_target_charset = $target_charset;
    }

    /**
     * Class destructor
     * 
     * @return void
     */
    function __destruct()
    {
        $this->clear();
    }

    /**
     * Load
     * 
     * @param string $str             HTML document string
     * @param bool   $lowercase       Parse tags in lowercase
     * @param bool   $stripRN         Strip newlines
     * @param string $defaultBRText   Default text for <br> elements
     * @param string $defaultSpanText Default text for <span> elements
     * @param int    $options         Additional options
     * 
     * @return object
     */
    function load(
        $str,
        $lowercase = true,
        $stripRN = true,
        $defaultBRText = DEFAULT_BR_TEXT,
        $defaultSpanText = DEFAULT_SPAN_TEXT,
        $options = 0
    ) {
        global $debug_object;

        // prepare
        $this->prepare($str, $lowercase, $defaultBRText, $defaultSpanText);

        // Per sourceforge http://sourceforge.net/tracker/?func=detail&aid=2949097&group_id=218559&atid=1044037
        // Script tags removal now preceeds style tag removal.
        // strip out <script> tags
        $this->removeNoise("'<\s*script[^>]*[^/]>(.*?)<\s*/\s*script\s*>'is");
        $this->removeNoise("'<\s*script\s*>(.*?)<\s*/\s*script\s*>'is");

        // strip out the \r \n's if we are told to.
        if ($stripRN) {
            $this->doc = str_replace("\r", ' ', $this->doc);
            $this->doc = str_replace("\n", ' ', $this->doc);

            // set the length of content since we have changed it.
            $this->size = strlen($this->doc);
        }

        // strip out cdata
        $this->removeNoise("'<!\[CDATA\[(.*?)\]\]>'is", true);
        // strip out comments
        $this->removeNoise("'<!--(.*?)-->'is");
        // strip out <style> tags
        $this->removeNoise("'<\s*style[^>]*[^/]>(.*?)<\s*/\s*style\s*>'is");
        $this->removeNoise("'<\s*style\s*>(.*?)<\s*/\s*style\s*>'is");
        // strip out preformatted tags
        $this->removeNoise("'<\s*(?:code)[^>]*>(.*?)<\s*/\s*(?:code)\s*>'is");
        // strip out server side scripts
        $this->removeNoise("'(<\?)(.*?)(\?>)'s", true);

        if ($options & HDOM_SMARTY_AS_TEXT) { // Strip Smarty scripts
            $this->removeNoise("'(\{\w)(.*?)(\})'s", true);
        }

        // parsing
        $this->parse();
        // end
        $this->root->p_[HDOM_INFO_END] = $this->cursor;
        $this->parseCharset();

        // make load function chainable
        return $this;
    }

    /**
     * Load file
     * 
     * @return object
     */
    function loadFile()
    {
        $args = func_get_args();

        if (($doc = call_user_func_array('file_get_contents', $args)) !== false) {
            $this->load($doc, true);
        } else {
            return false;
        }
    }

    /**
     * Set callback
     * 
     * @param string $function_name Function name
     * 
     * @return void
     */
    function setCallback($function_name)
    {
        $this->callback = $function_name;
    }

    /**
     * Remove callback
     * 
     * @return void
     */
    function removeCallback()
    {
        $this->callback = null;
    }

    /**
     * Save
     * 
     * @param string $filepath File path
     * 
     * @return string
     */
    function save($filepath = '')
    {
        $ret = $this->root->innertext();
        if ($filepath !== '') {
            file_put_contents($filepath, $ret, LOCK_EX); 
        }
        return $ret;
    }

    /**
     * Find
     * 
     * @param string $selector  CSS style selector
     * @param int    $idx       Element index
     * @param bool   $lowercase Match tag names case insensitive
     * 
     * @return mixed
     */
    function find($selector, $idx = null, $lowercase = false)
    {
        return $this->root->find($selector, $idx, $lowercase);
    }

    /**
     * Clear
     * 
     * @return void
     */
    function clear()
    {
        if (isset($this->nodes)) {
            foreach ($this->nodes as $n) {
                $n->clear();
                $n = null;
            }
        }

        // This add next line is documented in the sourceforge repository.
        // 2977248 as a fix for ongoing memory leaks that occur even with the
        // use of clear.
        if (isset($this->children)) {
            foreach ($this->children as $n) {
                $n->clear();
                $n = null;
            }
        }

        if (isset($this->parent)) {
            $this->parent->clear();
            unset($this->parent);
        }

        if (isset($this->root)) {
            $this->root->clear();
            unset($this->root);
        }

        unset($this->doc);
        unset($this->noise);
    }

    /**
     * Dump
     * 
     * @param bool $show_attr Attributes are included
     * 
     * @return string
     */
    function dump($show_attr = true)
    {
        $this->root->dump($show_attr);
    }

    /**
     * Prepare
     * 
     * @param string $str             HTML document string
     * @param bool   $lowercase       Parse in lowercase
     * @param string $defaultBRText   Default text for <be> elements
     * @param string $defaultSpanText Default test for <span> elements
     * 
     * @return void
     */
    protected function prepare(
        $str, $lowercase = true,
        $defaultBRText = DEFAULT_BR_TEXT,
        $defaultSpanText = DEFAULT_SPAN_TEXT
    ) {
        $this->clear();

        $this->doc = trim($str);
        $this->size = strlen($this->doc);
        $this->original_size = $this->size; // original size of the html
        $this->pos = 0;
        $this->cursor = 1;
        $this->noise = array();
        $this->nodes = array();
        $this->lowercase = $lowercase;
        $this->default_br_text = $defaultBRText;
        $this->default_span_text = $defaultSpanText;
        $this->root = new SimpleHtmlDomNode($this);
        $this->root->tag = 'root';
        $this->root->p_[HDOM_INFO_BEGIN] = -1;
        $this->root->nodetype = HDOM_TYPE_ROOT;
        $this->parent = $this->root;
        if ($this->size > 0) {
            $this->char = $this->doc[0]; 
        }
    }

    /**
     * Parse
     * 
     * @return void
     */
    protected function parse()
    {
        while (true) {
            // Read next tag if there is no text between current position and the
            // next opening tag.
            if (($s = $this->copyUntilChar('<')) === '') {
                if ($this->readTag()) {
                    continue;
                } else {
                    return true;
                }
            }

            // Add a text node for text between tags
            $node = new SimpleHtmlDomNode($this);
            ++$this->cursor;
            $node->p_[HDOM_INFO_TEXT] = $s;
            $this->linkNodes($node, false);
        }
    }

    /**
     * Parse charset
     * 
     * @return void
     */
    protected function parseCharset()
    {
        global $debug_object;

        $charset = null;

        if (function_exists('get_last_retrieve_url_contents_content_type')) {
            $contentTypeHeader = get_last_retrieve_url_contents_content_type();
            $success = preg_match('/charset=(.+)/', $contentTypeHeader, $matches);
            if ($success) {
                $charset = $matches[1];
                if (is_object($debug_object)) {
                    $debug_object->debug_log(
                        2,
                        'header content-type found charset of: '
                        . $charset
                    );
                }
            }
        }

        if (empty($charset)) {
            // https://www.w3.org/TR/html/document-metadata.html#statedef-http-equiv-content-type
            $el = $this->root->find('meta[http-equiv=Content-Type]', 0, true);

            if (!empty($el)) {
                $fullvalue = $el->content;
                if (is_object($debug_object)) {
                    $debug_object->debug_log(
                        2,
                        'meta content-type tag found'
                        . $fullvalue
                    );
                }

                if (!empty($fullvalue)) {
                    $success = preg_match(
                        '/charset=(.+)/i',
                        $fullvalue,
                        $matches
                    );

                    if ($success) {
                              $charset = $matches[1];
                    } else {
                        // If there is a meta tag, and they don't specify the
                        // character set, research says that it's typically
                        // ISO-8859-1
                        if (is_object($debug_object)) {
                            $debug_object->debug_log(
                                2,
                                'meta content-type tag couldn\'t be parsed.'
                                .' using iso-8859 default.'
                            );
                        }

                        $charset = 'ISO-8859-1';
                    }
                }
            }
        }

        if (empty($charset)) {
            // https://www.w3.org/TR/html/document-metadata.html#character-encoding-declaration
            if ($meta = $this->root->find('meta[charset]', 0)) {
                $charset = $meta->charset;
                if (is_object($debug_object)) {
                    $debug_object->debug_log(2, 'meta charset: ' . $charset);
                }
            }
        }

        if (empty($charset)) {
            // Try to guess the charset based on the content
            // Requires Multibyte String (mbstring) support (optional)
            if (function_exists('mb_detect_encoding')) {
                /**
                 * Check for presence of mb_detect_encoding()
                 * mb_detect_encoding() is not intended to distinguish between
                 * charsets, especially single-byte charsets. Its primary
                 * purpose is to detect which multibyte encoding is in use,
                 * i.e. UTF-8, UTF-16, shift-JIS, etc.
                 *
                 * -- https://bugs.php.net/bug.php?id=38138
                 *
                 * Adding both CP1251/ISO-8859-5 and CP1252/ISO-8859-1 will
                 * always result in CP1251/ISO-8859-5 and vice versa.
                 *
                 * Thus, only detect if it's either UTF-8 or CP1252/ISO-8859-1
                 * to stay compatible.
                 */
                $encoding = mb_detect_encoding(
                    $this->doc,
                    array( 'UTF-8', 'CP1252', 'ISO-8859-1' )
                );

                if ($encoding === 'CP1252' || $encoding === 'ISO-8859-1') {
                       // Due to a limitation of mb_detect_encoding
                       // 'CP1251'/'ISO-8859-5' will be detected as
                       // 'CP1252'/'ISO-8859-1'. This will cause iconv to fail, in
                       // which case we can simply assume it is the other charset.
                    if (!@iconv('CP1252', 'UTF-8', $this->doc)) {
                        $encoding = 'CP1251';
                    }
                }

                if ($encoding !== false) {
                    $charset = $encoding;
                    if (is_object($debug_object)) {
                        $debug_object->debug_log(2, 'mb_detect: ' . $charset);
                    }
                }
            }
        }

        if (empty($charset)) {
            // Assume it's UTF-8 as it is the most likely charset to be used
            $charset = 'UTF-8';
            if (is_object($debug_object)) {
                $debug_object->debug_log(2, 'No match found, assume ' . $charset);
            }
        }

        // Since CP1252 is a superset, if we get one of it's subsets, we want
        // it instead.
        if ((strtolower($charset) == 'iso-8859-1')
            || (strtolower($charset) == 'latin1')
            || (strtolower($charset) == 'latin-1')
        ) {
            $charset = 'CP1252';
            if (is_object($debug_object)) {
                $debug_object->debug_log(
                    2,
                    'replacing ' . $charset . ' with CP1252 as its a superset'
                );
            }
        }

        if (is_object($debug_object)) {
            $debug_object->debug_log(1, 'EXIT - ' . $charset);
        }

        return $this->p_charset = $charset;
    }

    /**
     * Read tag
     * 
     * @return bool
     */
    protected function readTag()
    {
        // Set end position if no further tags found
        if ($this->char !== '<') {
            $this->root->p_[HDOM_INFO_END] = $this->cursor;
            return false;
        }

        $begin_tag_pos = $this->pos;
        $this->char = (++$this->pos < $this->size) ?
            $this->doc[$this->pos] : null; // next

        // end tag
        if ($this->char === '/') {
            $this->char = (++$this->pos < $this->size) ?
                $this->doc[$this->pos] : null; // next

            // Skip whitespace in end tags (i.e. in "</   html>")
            $this->skip($this->token_blank);
            $tag = $this->copyUntilChar('>');

            // Skip attributes in end tags
            if (($pos = strpos($tag, ' ')) !== false) {
                $tag = substr($tag, 0, $pos);
            }

            $parent_lower = strtolower($this->parent->tag);
            $tag_lower = strtolower($tag);

            // The end tag is supposed to close the parent tag. Handle situations
            // when it doesn't
            if ($parent_lower !== $tag_lower) {
                // Parent tag does not have to be closed necessarily
                // (optional closing tag)
                // Current tag is a block tag, so it may close an ancestor
                if (isset($this->optional_closing_tags[$parent_lower])
                    && isset($this->block_tags[$tag_lower])
                ) {

                    $this->parent->p_[HDOM_INFO_END] = 0;
                    $org_parent = $this->parent;

                    // Traverse ancestors to find a matching opening tag
                    // Stop at root node
                    while (
                        ($this->parent->parent) &&
                        strtolower($this->parent->tag) !== $tag_lower) {
                        $this->parent = $this->parent->parent;
                    }

                    // If we don't have a match add current tag as text node
                    if (strtolower($this->parent->tag) !== $tag_lower) {
                        $this->parent = $org_parent; // restore origonal parent

                        if ($this->parent->parent) {
                            $this->parent = $this->parent->parent;
                        }

                        $this->parent->p_[HDOM_INFO_END] = $this->cursor;
                        return $this->asTextNode($tag);
                    }
                } elseif (($this->parent->parent)
                    && isset($this->block_tags[$tag_lower])
                ) {
                    // Grandparent exists and current tag is a block tag, so our
                    // parent doesn't have an end tag
                    $this->parent->p_[HDOM_INFO_END] = 0; // No end tag
                    $org_parent = $this->parent;

                    // Traverse ancestors to find a matching opening tag
                    // Stop at root node
                    while (($this->parent->parent)
                    && strtolower($this->parent->tag) !== $tag_lower
                    ) {
                        $this->parent = $this->parent->parent;
                    }

                    // If we don't have a match add current tag as text node
                    if (strtolower($this->parent->tag) !== $tag_lower) {
                        $this->parent = $org_parent; // restore origonal parent
                        $this->parent->p_[HDOM_INFO_END] = $this->cursor;
                        return $this->asTextNode($tag);
                    }
                } elseif (($this->parent->parent)
                    && strtolower($this->parent->parent->tag) === $tag_lower
                ) { // Grandparent exists and current tag closes it
                    $this->parent->p_[HDOM_INFO_END] = 0;
                    $this->parent = $this->parent->parent;
                } else { // Random tag, add as text node
                    return $this->asTextNode($tag);
                }
            }

            // Set end position of parent tag to current cursor position
            $this->parent->p_[HDOM_INFO_END] = $this->cursor;

            if ($this->parent->parent) {
                $this->parent = $this->parent->parent;
            }

            $this->char = (++$this->pos < $this->size) ?
                $this->doc[$this->pos] : null; // next
            return true;
        }

        // start tag
        $node = new SimpleHtmlDomNode($this);
        $node->p_[HDOM_INFO_BEGIN] = $this->cursor;
        ++$this->cursor;
        $tag = $this->copyUntil($this->token_slash); // Get tag name
        $node->tag_start = $begin_tag_pos;

        // doctype, cdata & comments...
        // <!DOCTYPE html>
        // <![CDATA[ ... ]]>
        // <!-- Comment -->
        if (isset($tag[0]) && $tag[0] === '!') {
            $node->p_[HDOM_INFO_TEXT] = '<' . $tag . $this->copyUntilChar('>');

            if (isset($tag[2]) && $tag[1] === '-' && $tag[2] === '-') {
                // Comment ("<!--")
                $node->nodetype = HDOM_TYPE_COMMENT;
                $node->tag = 'comment';
            } else { // Could be doctype or CDATA but we don't care
                $node->nodetype = HDOM_TYPE_UNKNOWN;
                $node->tag = 'unknown';
            }

            if ($this->char === '>') {
                $node->p_[HDOM_INFO_TEXT] .= '>'; 
            }

            $this->linkNodes($node, true);
            $this->char = (++$this->pos < $this->size) ?
                $this->doc[$this->pos] : null; // next
            return true;
        }

        // The start tag cannot contain another start tag, if so add as text
        // i.e. "<<html>"
        if ($pos = strpos($tag, '<') !== false) {
            $tag = '<' . substr($tag, 0, -1);
            $node->p_[HDOM_INFO_TEXT] = $tag;
            $this->linkNodes($node, false);
            $this->char = $this->doc[--$this->pos]; // prev
            return true;
        }

        // Handle invalid tag names (i.e. "<html#doc>")
        if (!preg_match('/^\w[\w:-]*$/', $tag)) {
            $node->p_[HDOM_INFO_TEXT] = '<' . $tag . $this->copyUntil('<>');

            // Next char is the beginning of a new tag, don't touch it.
            if ($this->char === '<') {
                $this->linkNodes($node, false);
                return true;
            }

            // Next char closes current tag, add and be done with it.
            if ($this->char === '>') {
                $node->p_[HDOM_INFO_TEXT] .= '>'; 
            }
            $this->linkNodes($node, false);
            $this->char = (++$this->pos < $this->size) ?
                $this->doc[$this->pos] : null; // next
            return true;
        }

        // begin tag, add new node
        $node->nodetype = HDOM_TYPE_ELEMENT;
        $tag_lower = strtolower($tag);
        $node->tag = ($this->lowercase) ? $tag_lower : $tag;

        // handle optional closing tags
        if (isset($this->optional_closing_tags[$tag_lower])) {
            // Traverse ancestors to close all optional closing tags
            while (
                isset(
                    $this->optional_closing_tags[$tag_lower]
                    [strtolower($this->parent->tag)]
                )
            ) {
                $this->parent->p_[HDOM_INFO_END] = 0;
                $this->parent = $this->parent->parent;
            }
            $node->parent = $this->parent;
        }

        $guard = 0; // prevent infinity loop

        // [0] Space between tag and first attribute
        $space = array($this->copySkip($this->token_blank), '', '');

        // attributes
        do {
            // Everything until the first equal sign should be the attribute name
            $name = $this->copyUntil($this->token_equal);

            if ($name === '' && $this->char !== null && $space[0] === '') {
                break;
            }

            if ($guard === $this->pos) { // Escape infinite loop
                $this->char = (++$this->pos < $this->size) ?
                    $this->doc[$this->pos] : null; // next
                continue;
            }

            $guard = $this->pos;

            // handle endless '<'
            // Out of bounds before the tag ended
            if ($this->pos >= $this->size - 1 && $this->char !== '>') {
                $node->nodetype = HDOM_TYPE_TEXT;
                $node->p_[HDOM_INFO_END] = 0;
                $node->p_[HDOM_INFO_TEXT] = '<' . $tag . $space[0] . $name;
                $node->tag = 'text';
                $this->linkNodes($node, false);
                return true;
            }

            // handle mismatch '<'
            // Attributes cannot start after opening tag
            if ($this->doc[$this->pos - 1] == '<') {
                $node->nodetype = HDOM_TYPE_TEXT;
                $node->tag = 'text';
                $node->attr = array();
                $node->p_[HDOM_INFO_END] = 0;
                $node->p_[HDOM_INFO_TEXT] = substr(
                    $this->doc,
                    $begin_tag_pos,
                    $this->pos - $begin_tag_pos - 1
                );
                $this->pos -= 2;
                $this->char = (++$this->pos < $this->size) ?
                    $this->doc[$this->pos] : null; // next
                $this->linkNodes($node, false);
                return true;
            }

            if ($name !== '/' && $name !== '') { // this is a attribute name
                // [1] Whitespace after attribute name
                $space[1] = $this->copySkip($this->token_blank);

                $name = $this->restoreNoise($name); // might be a noisy name

                if ($this->lowercase) {
                    $name = strtolower($name); 
                }

                if ($this->char === '=') { // attribute with value
                    $this->char = (++$this->pos < $this->size) ?
                        $this->doc[$this->pos] : null; // next
                    $this->parseAttr($node, $name, $space); // get attribute value
                } else {
                    //no value attr: nowrap, checked selected...
                    $node->p_[HDOM_INFO_QUOTE][] = HDOM_QUOTE_NO;
                    $node->attr[$name] = true;
                    if ($this->char != '>') {
                        $this->char = $this->doc[--$this->pos]; 
                    } // prev
                }

                $node->p_[HDOM_INFO_SPACE][] = $space;

                // prepare for next attribute
                $space = array(
                $this->copySkip($this->token_blank),
                '',
                ''
                );
            } else { // no more attributes
                break;
            }
            // go until the tag ended
        } while ($this->char !== '>' && $this->char !== '/');

        $this->linkNodes($node, true);
        $node->p_[HDOM_INFO_ENDSPACE] = $space[0];

        // handle empty tags (i.e. "<div/>")
        if ($this->copyUntilChar('>') === '/') {
            $node->p_[HDOM_INFO_ENDSPACE] .= '/';
            $node->p_[HDOM_INFO_END] = 0;
        } else {
            // reset parent
            if (!isset($this->self_closing_tags[strtolower($node->tag)])) {
                $this->parent = $node;
            }
        }

        $this->char = (++$this->pos < $this->size) ?
            $this->doc[$this->pos] : null; // next

        // If it's a BR tag, we need to set it's text to the default text.
        // This way when we see it in plaintext, we can generate
        // formatting that the user wants.
        // since a br tag never has sub nodes, this works well.
        if ($node->tag === 'br') {
            $node->p_[HDOM_INFO_INNER] = $this->default_br_text;
        }

        return true;
    }

    /**
     * Parse attr
     * 
     * @param object $node  Current element
     * @param string $name  Attribute name
     * @param array  $space Array of whitespace
     * 
     * @return void
     */
    protected function parseAttr($node, $name, &$space)
    {
        $is_duplicate = isset($node->attr[$name]);

        if (!$is_duplicate) { // Copy whitespace between "=" and value
            $space[2] = $this->copySkip($this->token_blank);
        }

        switch ($this->char) {
        case '"':
            $quote_type = HDOM_QUOTE_DOUBLE;
            $this->char = (++$this->pos < $this->size) ?
                $this->doc[$this->pos] : null; // next
            $value = $this->copyUntilChar('"');
            $this->char = (++$this->pos < $this->size) ?
                $this->doc[$this->pos] : null; // next
            break;
        case '\'':
            $quote_type = HDOM_QUOTE_SINGLE;
            $this->char = (++$this->pos < $this->size) ?
                $this->doc[$this->pos] : null; // next
            $value = $this->copyUntilChar('\'');
            $this->char = (++$this->pos < $this->size) ?
                $this->doc[$this->pos] : null; // next
            break;
        default:
            $quote_type = HDOM_QUOTE_NO;
            $value = $this->copyUntil($this->token_attr);
        }

        $value = $this->restoreNoise($value);

        // PaperG: Attributes should not have \r or \n in them, that counts as
        // html whitespace.
        $value = str_replace("\r", '', $value);
        $value = str_replace("\n", '', $value);

        // PaperG: If this is a "class" selector, lets get rid of the preceeding
        // and trailing space since some people leave it in the multi class case.
        if ($name === 'class') {
            $value = trim($value);
        }

        if (!$is_duplicate) {
            $node->p_[HDOM_INFO_QUOTE][] = $quote_type;
            $node->attr[$name] = $value;
        }
    }

    /**
     * Link nodes
     * 
     * @param object $node     Node to link
     * @param bool   $is_child Makes node a sibling
     * 
     * @return void
     */
    protected function linkNodes(&$node, $is_child)
    {
        $node->parent = $this->parent;
        $this->parent->nodes[] = $node;
        if ($is_child) {
            $this->parent->children[] = $node;
        }
    }

    /**
     * As text node
     * 
     * @param string $tag Element tag name
     * 
     * @return bool
     */
    protected function asTextNode($tag)
    {
        $node = new SimpleHtmlDomNode($this);
        ++$this->cursor;
        $node->p_[HDOM_INFO_TEXT] = '</' . $tag . '>';
        $this->linkNodes($node, false);
        $this->char = (++$this->pos < $this->size) ?
            $this->doc[$this->pos] : null; // next
        return true;
    }

    /**
     * Skip
     * 
     * @param string $chars List of characters to skip
     * 
     * @return void
     */
    protected function skip($chars)
    {
        $this->pos += strspn($this->doc, $chars, $this->pos);
        $this->char = ($this->pos < $this->size) ?
            $this->doc[$this->pos] : null; // next
    }

    /**
     * Copy skip
     * 
     * @param string $chars List of characters to skip
     * 
     * @return void
     */
    protected function copySkip($chars)
    {
        $pos = $this->pos;
        $len = strspn($this->doc, $chars, $pos);
        $this->pos += $len;
        $this->char = ($this->pos < $this->size) ?
            $this->doc[$this->pos] : null; // next
        if ($len === 0) {
            return ''; 
        }
        return substr($this->doc, $pos, $len);
    }

    /**
     * Copy until
     * 
     * @param string $chars List of characters to stop copy at
     * 
     * @return void
     */
    protected function copyUntil($chars)
    {
        $pos = $this->pos;
        $len = strcspn($this->doc, $chars, $pos);
        $this->pos += $len;
        $this->char = ($this->pos < $this->size) ?
            $this->doc[$this->pos] : null; // next
        return substr($this->doc, $pos, $len);
    }

    /**
     * Copy until char
     * 
     * @param string $char Characters to stop copy at
     * 
     * @return void
     */
    protected function copyUntilChar($char)
    {
        if ($this->char === null) {
            return ''; 
        }

        if (($pos = strpos($this->doc, $char, $this->pos)) === false) {
            $ret = substr($this->doc, $this->pos, $this->size - $this->pos);
            $this->char = null;
            $this->pos = $this->size;
            return $ret;
        }

        if ($pos === $this->pos) {
            return ''; 
        }

        $pos_old = $this->pos;
        $this->char = $this->doc[$pos];
        $this->pos = $pos;
        return substr($this->doc, $pos_old, $pos - $pos_old);
    }

    /**
     * Remove noise
     * 
     * @param string $pattern    Regular expression matching noise
     * @param bool   $remove_tag Remove entire match
     * 
     * @return void
     */
    protected function removeNoise($pattern, $remove_tag = false)
    {
        global $debug_object;
        if (is_object($debug_object)) {
            $debug_object->debug_log_entry(1); 
        }

        $count = preg_match_all(
            $pattern,
            $this->doc,
            $matches,
            PREG_SET_ORDER | PREG_OFFSET_CAPTURE
        );

        for ($i = $count - 1; $i > -1; --$i) {
            $key = '___noise___' . sprintf('% 5d', count($this->noise) + 1000);

            if (is_object($debug_object)) {
                $debug_object->debug_log(2, 'key is: ' . $key);
            }

            $idx = ($remove_tag) ? 0 : 1; // 0 = entire match, 1 = submatch
            $this->noise[$key] = $matches[$i][$idx][0];
            $this->doc = substr_replace(
                $this->doc, $key, $matches[$i][$idx][1],
                strlen($matches[$i][$idx][0])
            );
        }

        // reset the length of content
        $this->size = strlen($this->doc);

        if ($this->size > 0) {
            $this->char = $this->doc[0];
        }
    }

    /**
     * Restore noise
     * 
     * @param string $text String containing noise placeholders
     * 
     * @return string
     */
    function restoreNoise($text)
    {
        global $debug_object;
        if (is_object($debug_object)) {
            $debug_object->debug_log_entry(1); 
        }

        while (($pos = strpos($text, '___noise___')) !== false) {
            // Sometimes there is a broken piece of markup, and we don't GET the
            // pos+11 etc... token which indicates a problem outside of us...

            // todo: "___noise___1000" (or any number with four or more digits)
            // in the DOM causes an infinite loop which could be utilized by
            // malicious software
            if (strlen($text) > $pos + 15) {
                $key = '___noise___'
                . $text[$pos + 11]
                . $text[$pos + 12]
                . $text[$pos + 13]
                . $text[$pos + 14]
                . $text[$pos + 15];

                if (is_object($debug_object)) {
                    $debug_object->debug_log(2, 'located key of: ' . $key);
                }

                if (isset($this->noise[$key])) {
                    $text = substr($text, 0, $pos)
                    . $this->noise[$key]
                    . substr($text, $pos + 16);
                } else {
                    // do this to prevent an infinite loop.
                    $text = substr($text, 0, $pos)
                    . 'UNDEFINED NOISE FOR KEY: '
                    . $key
                    . substr($text, $pos + 16);
                }
            } else {
                // There is no valid key being given back to us... We must get
                // rid of the ___noise___ or we will have a problem.
                $text = substr($text, 0, $pos)
                . 'NO NUMERIC NOISE KEY'
                . substr($text, $pos + 11);
            }
        }
        return $text;
    }

    /**
     * Search noise
     * 
     * @param string $text Noise place holder to find
     * 
     * @return string
     */
    function searchNoise($text)
    {
        global $debug_object;
        if (is_object($debug_object)) { 
            $debug_object->debug_log_entry(1); 
        }

        foreach ($this->noise as $noiseElement) {
            if (strpos($noiseElement, $text) !== false) {
                return $noiseElement;
            }
        }
    }

    /**
     * To string
     * 
     * @return string
     */
    function __toString()
    {
        return $this->root->innertext();
    }

    /**
     * Get
     * 
     * @param string $name Name
     * 
     * @return mixed
     */
    function __get($name)
    {
        switch ($name) {
        case 'outertext':
            return $this->root->innertext();
        case 'innertext':
            return $this->root->innertext();
        case 'plaintext':
            return $this->root->text();
        case 'charset':
            return $this->p_charset;
        case 'target_charset':
            return $this->p_target_charset;
        }
    }

    /**
     * Child nodes
     * 
     * @param int $idx Index of child element
     * 
     * @return mixed
     */
    function childNodes($idx = -1)
    {
        return $this->root->childNodes($idx);
    }

    /**
     * First child
     * 
     * @return object
     */
    function firstChild()
    {
        return $this->root->firstChild();
    }

    /**
     * Last child
     * 
     * @return object
     */
    function lastChild()
    {
        return $this->root->lastChild();
    }

    /**
     * Create e3lement
     * 
     * @param string $name  Name of element
     * @param string $value Value of element
     * 
     * @return object
     */
    function createElement($name, $value = null)
    {
        return @strGetHtml("<$name>$value</$name>")->firstChild();
    }

    /**
     * Create text node
     * 
     * @param string $value Value of new element
     * 
     * @return object
     */
    function createTextNode($value)
    {
        return @end(strGetHtml($value)->nodes);
    }

    /**
     * Get element by id
     * 
     * @param string $id Id of element to find
     * 
     * @return object
     */
    function getElementById($id)
    {
        return $this->find("#$id", 0);
    }

    /**
     * Get elements by id
     * 
     * @param string $id  Id of element to find
     * @param int    $idx Index of element
     * 
     * @return object
     */
    function getElementsById($id, $idx = null)
    {
        return $this->find("#$id", $idx);
    }

    /**
     * Get element by tag name
     * 
     * @param string $name Tag name of element
     *
     * @return object 
     */
    function getElementByTagName($name)
    {
        return $this->find($name, 0);
    }

    /**
     * Get elements by tag name
     * 
     * @param string $name Tag name of element
     * @param int    $idx  Index of elemt
     * 
     * @return object
     */
    function getElementsByTagName($name, $idx = -1)
    {
        return $this->find($name, $idx);
    }
}
